part of validation;

class Failure<F, S> extends Validation<F, S> {

  final F _data;

  /**
   * Const constructor for const initialization.
   */
  const Failure(F this._data);

  /**
   * An approximation of pattern matching on the Validation type.
   *
   * Always calls the supplied failure function with the inner wrapped
   * value of this Failure instance.
   */
  dynamic cata(dynamic failure(F fail), dynamic successful(S success)) =>
    failure(_data);

  /**
   * An optional failure projection of this Validation type.
   *
   * Returns the inner wrapped value of this Validation inside of a Some<F>.
   */
  Option<F> failure() => new Some(_data);

  /**
   * An optional success projection of this Validation type.
   *
   * Returns the const instance of None<S>.
   */
  Option<S> successful() => const None();

  /**
   * Attempts to unwrap the successful case with a fallback computation.
   *
   * Returns the result of calling the supplied alternative.
   */
  S getOrElse(S alternative()) => alternative();

  /**
   * Attempts to unwrap the successful case with a fallback value.
   *
   * Returns the supplied alternative.
   */
  S getOrDefault(S def) => def;

  /**
   * Computes an alternative for the failure case.
   *
   * Returns the result of calling the supplied alternative computation.
   */
  Validation<F, S> orElse(Validation<F,S> alternative()) => alternative();

  /**
   * Performs a map on this validation type.
   *
   * Returns itself immediately.
   */
  Validation<F, dynamic> map(dynamic mapper(S n)) => this;

  /**
   * Applies another applicative Validation to this one.
   *
   * Returns this immediately.
   */
  Validation<F, dynamic> ap(Validation<F, dynamic> other) => this;

  /**
   * Performs a monadic expansion on the validation type.
   *
   * Returns itself immediately.
   */
  Validation<F, dynamic> expand(Validation<F, dynamic> flatMapper(S n)) =>
    this;

  /**
   * Performs a map on the failure case.
   *
   * Returns a new Failure containing the result of calling the mapper with the
   * inner wrapped value.
   */
  Validation<dynamic, S> leftMap(dynamic mapper(F f)) =>
    new Failure(mapper(_data));

}